"""

"""
import numpy as np
import pandas as pd
import networkx as nx
import multiprocessing
from multiprocessing import Queue, JoinableQueue, Process
import inspect
from navipy.database import DataBaseLoad
import navipy.moving.maths as navimomath


def defaultcallback(*args, **kwargs):
    """default call back"""
    raise NameError('No Callback')


class AbstractAgent():
    def __init__(self):
        self._sensors = defaultcallback
        self._motion = defaultcallback
        self._alter_posorientvel = defaultcallback
        self._posorient_col = ['x', 'y', 'z',
                               'alpha_0', 'alpha_1', 'alpha_2']
        self._velocity_col = ['d' + col for col in self._posorient_col]
        self._posorient_vel_col = self._posorient_col.copy()
        self._posorient_vel_col.extend(self._velocity_col)
        self._posorient_vel = pd.Series(
            index=self._posorient_vel_col,
            data=np.nan)

    @property
    def posorient(self):
        return self._posorient_vel.loc[self._posorient_col].copy()

    @posorient.setter
    def posorient(self, posorient):
        if isinstance(posorient, pd.Series) is False:
            raise TypeError('posorient should be a pandas Series')
        for col in self._posorient_col:
            if col not in posorient.index:
                raise KeyError(
                    'posorient should have {} as index'.format(col))
        self._posorient_vel.loc[self._posorient_col] = \
            posorient.loc[self._posorient_col]

    @property
    def velocity(self):
        return self._posorient_vel.loc[self._velocity_col].copy()

    @velocity.setter
    def velocity(self, velocity):
        if isinstance(velocity, pd.Series) is False:
            raise TypeError('velocity should be a pandas Series')
        for col in self._velocity_col:
            if col not in velocity.index:
                raise KeyError(
                    'velocity should have {} as index'.format(col))
        self._posorient_vel.loc[self._velocity_col] = \
            velocity.loc[self._velocity_col]

    @property
    def motion(self):
        return inspect.getsourcelines(self._motion)

    @property
    def sensors(self):
        return inspect.getsourcelines(self._sensors)

    @property
    def alter_posorientvel(self):
        return inspect.getsourcelines(self._alter_posorientvel)

    def move(self):
        scene = self._sensors(self.posorient)
        newpos = self._motion(self.posorient, scene)
        alteredpos = self._alter_posorientvel(newpos)
        self.posorient = alteredpos
        self.velocity = alteredpos

    def fly(self, max_nstep, return_tra=False):
        """move cyberbee until max step has been performed
        """
        if return_tra:
            trajectory = pd.DataFrame(index=range(0, max_nstep),
                                      columns=self.posorient_vel_col)
            trajectory.loc[0, :] = self._posorient_vel.copy()
        for stepi in range(1, max_nstep):
            self.move()
            if return_tra:
                trajectory.loc[stepi, :] = self._posorient_vel.copy()
        if return_tra:
            return trajectory
        else:
            return None


class CyberBeeAgent(AbstractAgent):
    """
    A common method to make an agent moves is to update the sensory \
    information at the current agent location, then process this \
    information such to deduce the agent motion, and finally displaced\
    the agent at its new location. This iterative method can be used \
    with a wide range of models.
    In navipy the update of the sensory information is done by the \
    Cyberbee a class interfacing blender with the navipy, such that \
    visual information can be rendered at the agent location.
    To use the CyberBeeAgent you first need to create a function with \
    input the scene and the position orientation of the agent, and \
    with output the agent motion. The CyberBeeAgent can then be moved \
    for a single step, or fly until a given number of movement has \
    been effectuated or the agent stopped.
    """

    def __init__(self, cyberbee):
        AbstractAgent.__init__(self)
        AbstractAgent._alter_posorientvel = \
            lambda motion_vec: navimomath.next_pos(motion_vec,
                                                   move_mode='free_run')
        self.sensors = cyberbee.scene

    @AbstractAgent.sensors.setter
    def sensors(self, cyberbee):
        self._sensors = cyberbee.scene

    @AbstractAgent.motion.setter
    def motion(self, motion):
        self._motion = motion


class GridAgent(AbstractAgent, Process):
    """
    The use of a close loop model including visual rendering is \
    sometimes too slow to efficiently test several models or tune the \
    parameters of a given models. The GridAgent solves this problem by \
    restricting the agent motion on locations with rendered scene. The \
    agent moves thus on a grid, and its next position is always \
    snapped to the closest grid location. The finer the grid is, the \
    larger the database storing all sceneries and grid location is; \
    but also the more accurate the agent motion is. The grid size \
    depend on the storage space, the time you can wait for the \
    database creation, and how sensitive to exact location your model is.
    Similar to the CyberBeeAgent, your navigational model should be \
    contained in a function with input the scene and the position \
    orientation of the agent, and with output the agent motion. The \
    agent can be move a single step, or fly until a given number of \
    movement has been effectuated or the agent stopped. It is here \
    worth mentioning that the GridAgent inherit from the Process \
    class of the multiprocessing module of the standard python \
    library. Thus, several GridAgents can safely be run in parallel.
    """

    def __init__(self, database_filename,
                 posorients_queue=None,
                 results_queue=None):
        if (posorients_queue is not None) and (results_queue is not None):
            multiprocessing.Process.__init__(self)
        AbstractAgent.__init__(self)
        self._alter_posorientvel = self.snap_to_grid
        self.sensors = database_filename

    @AbstractAgent.sensors.setter
    def sensors(self, database_filename):
        self.db = DataBaseLoad(database_filename)
        self._posorients = self.db.posorients
        self._sensors = self.db.scene

    @AbstractAgent.motion.setter
    def motion(self, motion):
        self._motion = motion

    @property
    def mode_of_motion(self):
        """
        """
        toreturn = self._mode_move
        toreturn['describe'] = \
            navimomath.mode_moves_supported()[
                self._mode_move['mode']]['describe']
        return toreturn

    @mode_of_motion.setter
    def mode_of_motion(self, mode):
        """

        """
        if not isinstance(mode, dict):
            raise TypeError('Mode is not a dictionary')
        if 'mode' not in mode:
            raise KeyError("'mode' is not a key of mode")
        if 'param' not in mode:
            raise KeyError("'param' is not a key of mode")
        if mode['mode'] in navimomath.mode_moves_supported().keys():
            for param in navimomath.mode_moves_supported()[
                    mode['mode']]['param']:
                if param not in mode['param']:
                    raise KeyError(
                        "'{}' is not in mode['param']".format(param))
            self._mode_move = mode
        else:
            raise ValueError('mode is not supported')

    def snap_to_grid(self, posorient_vel):
        posorient_vel = navimomath.next_pos(
            posorient_vel,
            move_mode=self._mode_move['mode'],
            move_param=self._mode_move['param'])
        tmp = navimomath.closest_pos(
            posorient_vel, self._posorients)
        posorient_vel.loc[self._posorient_col] = \
            tmp.loc[self._posorient_col]
        posorient_vel.name = tmp.name
        return posorient_vel

    def move(self):
        if hasattr(self, '_mode_move'):
            AbstractAgent.move(self)
        else:
            raise AttributeError(
                'GridAgent object has no attribute _mode_move\n' +
                'Please set the mode of motion')

    def fly(self, max_nstep, return_tra=False):
        if hasattr(self, '_mode_move'):
            return AbstractAgent.fly(self, max_nstep, return_tra)
        else:
            raise AttributeError(
                'GridAgent object has no attribute _mode_move\n' +
                'Please set the mode of motion')

    def run(self):
        """ Only supported when multiprocess"""
        if self._posorients_queue is None or self._results_queue is None:
            raise NameError('Single agent class has not be inititialised '
                            + 'with multiprocessing suppport')
        proc_name = self.name
        print('Process {} started'.format(proc_name))
        while True:
            start_posorient = self._posorients_queue.get(timeout=1)
            if start_posorient is None:
                # Poison pill means shutdown)
                break
            common_id = list(set(start_posorient.index).intersection(
                self._posorientvel.index))
            self._posorientvel.loc[common_id] = start_posorient.loc[common_id]
            self.move()
            next_posorient = self._posorientvel

            self._posorients_queue.task_done()
            self._results_queue.put((start_posorient, next_posorient))
        self._posorients_queue.task_done()
        print('Process {} done'.format(proc_name))


class GraphAgent():
    """
    As mentioned above, in every model of navigation the agent motion \
    is derived from its current external state, its position \
    orientation as well as the derivatives, and its internal state. \
    However, when the agent motion is only derived from its current \
    position orientation, and what is seen from this location, the \
    simulation of an agent can be drastically simplified. Indeed, \
    not only the scene at relevant location can be pre-rendered, but \
    the motion of the agent from those locations as well. The agent \
    being restricted to move from relevant locations to relevant \
    locations, a graph of interconnected locations can be built. The \
    nodes of the graph are the relevant locations, and the directed \
    edges the motion of the agent from one location to the next. \
    GraphAgent can build such graph by simply using a database of \
    pre-rendered scenery at relevant locations, and a function \
    giving the motion of the agent from a scene and the agent \
    position orientation. Once the graph has been generated, \
    attractors can be found, the number of locations converging to \
    those (i.e. the catchment area or volume), if two locations are \
    connected, etc.
    To speed up certain calculations, additional values can stored \
    at each graph node and access from the callback function. It is \
    worth mentioning a warning here. The size of the graph can be \
    incredibly large. Thus, not too much information can be stored \
    at each node. To assess the memory size of the graph before \
    creating it, one can use the tool agent.tools.assess_graphmemsize.
    """

    def __init__(self, database_filename):
        self.db = DataBaseLoad(database_filename)
        # Init the graph
        self._graph = nx.DiGraph()
        for row_id, posor in self.db.posorients.iterrows():
            posor.name = row_id
            self._graph.add_node(row_id,
                                 posorient=posor)

    @property
    def graph(self):
        return self._graph

    @graph.setter
    def graph(self, graph):
        if isinstance(graph, nx.DiGraph) is False:
            raise TypeError('graph is not a nx.DiGraph')
        self._graph = graph.copy()
        self.check_graph()

    def build_graph(self, callback_function,
                    ncpu=5,
                    timeout=1):
        # Build a list of nodes
        results_edges = []
        posorients_queue = JoinableQueue()
        results_queue = Queue()
        for node in self._graph.nodes:
            posorients_queue.put(self._graph.nodes[node]['posorient'])

        # Start ndatabase loader
        num_agents = ncpu
        agents = [GridAgent(self.dbname,
                            posorients_queue=posorients_queue,
                            results_queue=results_queue)
                  for _ in range(num_agents)]
        for w in agents:
            w.callback_function = callback_function
            w.start()

        # Add a poison pill for each agent
        for _ in range(num_agents):
            posorients_queue.put(None)

        # Wait for all of the tasks to finish
        # posorients_queue.join()

        for _ in range(nx.number_of_nodes(self._graph)):
            result = results_queue.get(timeout=timeout)
            results_edges.append((result[0].name,
                                  result[1].name))
            # print(results_edges[-1])
        self._graph.add_edges_from(results_edges)
        self.check_graph()

    def check_graph(self):
        self.check_single_target()

    def check_single_target(self):
        for node in self._graph.nodes:
            # not connected -> loop not ran
            for count, _ in enumerate(self._graph.neighbors(node)):
                # count == 0 -> connected to one node
                # count == 1 -> connected to two nodes
                if count > 0:
                    raise ValueError(
                        'Node {} leads to several locations'.format(node))

    def find_attractors(self):
        """Return a list of node going to each attractor in a graph
        """
        attractors = list()
        for attractor in nx.attracting_components(self._graph):
            att = dict()
            att['attractor'] = attractor
            attractors.append(att)
        return attractors

    def find_attractors_sources(self, attractors=None):
        """Find all sources going to each attractors
        """
        if attractors is None:
            attractors = self.find_attractors()

        if isinstance(attractors, list) is False:
            raise TypeError('Attractors should be a list of dict')
        elif len(attractors) == 0:
            raise ValueError('No attractors found')

        # Check attractor
        for att in attractors:
            keyatt = att.keys()
            if 'attractor' not in keyatt:
                raise ValueError(
                    'Each attractors should contain the key attractor')

        # Calculate connection
        for att_i, att in enumerate(attractors):

            # [0] because one node of the attractor is enough
            # all other node of the attractor are connected to this one
            target = list(att['attractor'])[0]
            attractors[att_i]['paths'] = \
                nx.shortest_path(self.graph, target=target)
            attractors[att_i]['sources'] = \
                list(attractors[att_i]['paths'].keys())
        return attractors

    def catchment_area(self, attractors=None):
        """Return the catchment area for attractors
        """
        if attractors is None:
            attractors = self.find_attractors_sources()

        if isinstance(attractors, list) is False:
            raise TypeError('Attractors should be a list of dict')
        elif len(attractors) == 0:
            raise ValueError('No attractors found')

        # Check attractor
        for att in attractors:
            keyatt = att.keys()
            if 'sources' not in keyatt:
                raise ValueError(
                    'Each attractors should contains a list of sources')

        return [len(att['sources']) for att in attractors]

    def reach_goals(self, goals):
        """ Return all paths to the goals """
        return nx.shortest_path(self._graph, target=goals)

    def neighboring_nodes(self, target):
        """ Return the nodes going to the target """
        # Reverse graph because nx.neighbors give the end node
        # and we want to find the start node going to target
        # not where target goes.
        tmpgraph = self._graph.reverse(copy=True)
        neighbors = tmpgraph.neighbors(target)
        return neighbors
