"""
Test of agent
"""
import numpy as np
import pandas as pd
import networkx as nx
import navipy.moving.agent as naviagent
import navipy.database as navidb
import pkg_resources

import unittest


class TestNavipyMovingAgent(unittest.TestCase):

    def setUp(self):
        self.mydb_filename = pkg_resources.resource_filename(
            'navipy', 'resources/database.db')
        self.mydb = navidb.DataBaseLoad(self.mydb_filename)
        self.__posorient_col = ['x', 'y', 'z',
                                'alpha_0', 'alpha_1', 'alpha_2']
        self.__velocity_col = ['d' + col for col in self.__posorient_col]
        self.__posorient_vel_col = self.__posorient_col
        self.__posorient_vel_col.extend(self.__velocity_col)

    #
    # AbstractAgent
    #
    def test_move_abstractagent(self):
        agent = naviagent.AbstractAgent()
        with self.assertRaises(NameError):
            agent.move()

    def test_fly_abstractagent(self):
        agent = naviagent.AbstractAgent()
        with self.assertRaises(NameError):
            agent.fly(max_nstep=10)

    #
    # GridAgent
    #
    def test_move_gridagent(self):
        agent = naviagent.GridAgent(self.mydb_filename)
        initposorient = agent.db.posorients.loc[13, :]
        initposovel = pd.Series(data=0,
                                index=self.__posorient_vel_col)
        initposovel.loc[initposorient.index] = initposorient
        agent.posorient = initposovel
        agent.motion = lambda posorient, scene:\
            pd.Series(data=0,
                      index=self.__posorient_vel_col)
        with self.assertRaises(AttributeError):
            agent.move()
        mode_move = {'mode': 'on_cubic_grid',
                     'param': {'grid_spacing':
                               pd.Series(data=1,
                                         index=['dx', 'dy', 'dz'])}}
        agent.mode_of_motion = mode_move
        agent.move()
        obtained = agent.posorient
        self.assertTrue(np.allclose(
            obtained, initposorient.loc[obtained.index]))

    def test_fly_gridagent(self):
        agent = naviagent.GridAgent(self.mydb_filename)
        initposorient = agent.db.posorients.loc[13, :]
        initposovel = pd.Series(data=0,
                                index=self.__posorient_vel_col)
        initposovel.loc[initposorient.index] = initposorient
        agent.posorient = initposovel
        agent.motion = lambda posorient, scene:\
            pd.Series(data=0,
                      index=self.__posorient_vel_col)
        with self.assertRaises(AttributeError):
            agent.fly(max_nstep=10)
        mode_move = {'mode': 'on_cubic_grid',
                     'param': {'grid_spacing':
                               pd.Series(data=1,
                                         index=['dx', 'dy', 'dz'])}}
        agent.mode_of_motion = mode_move
        agent.fly(max_nstep=10)
        obtained = agent.posorient
        self.assertTrue(np.allclose(obtained,
                                    initposorient.loc[obtained.index]))

    #
    # GraphAgent
    #

    def test_init_graphagent(self):
        agent = naviagent.GraphAgent(self.mydb_filename)
        self.assertEqual(sorted(agent.graph.nodes),
                         sorted(list(self.mydb.posorients.index)),
                         'Init of graph failed. Node missmatch')

    def test_graph_setter(self):
        agent = naviagent.GraphAgent(self.mydb_filename)
        graph_nodes = list(agent.graph.nodes)
        graph_edges = list()
        for gnode in graph_nodes[1:]:
            graph_edges.append((gnode, graph_nodes[0]))

        graph = nx.DiGraph()
        graph.add_nodes_from(graph_nodes)
        graph.add_edges_from(graph_edges)
        agent.graph = graph

        graph_edges.append((graph_nodes[2], graph_nodes[1]))
        graph = nx.DiGraph()
        graph.add_nodes_from(graph_nodes)
        graph.add_edges_from(graph_edges)
        with self.assertRaises(ValueError):
            agent.graph = graph

    def test_catchment_area(self):
        """
        1 Test all node to first
        2 Test 11 nodes to first, 14 to 12th
        3 Two loops attractors
        """
        # Test all node to first
        agent = naviagent.GraphAgent(self.mydb_filename)

        graph_nodes = list(agent.graph.nodes)
        graph_edges = list()
        for gnode in graph_nodes[1:]:
            graph_edges.append((gnode, graph_nodes[0]))

        graph = nx.DiGraph()
        graph.add_nodes_from(graph_nodes)
        graph.add_edges_from(graph_edges)
        agent.graph = graph
        attractors = agent.find_attractors()
        self.assertEqual(len(attractors), 1, 'Too many or too few attractors')
        attractors = agent.find_attractors_sources(attractors)
        catchment_area = agent.catchment_area(attractors)
        self.assertEqual(catchment_area, [len(graph_nodes)],
                         'Too big or too short catchment area')

        # Test 11 nodes to first, 14 to 12th
        graph_edges = list()
        for gnode in graph_nodes[1:11]:
            graph_edges.append((gnode, graph_nodes[0]))
        for gnode in graph_nodes[11:]:
            graph_edges.append((gnode, graph_nodes[11]))

        graph = nx.DiGraph()
        graph.add_nodes_from(graph_nodes)
        graph.add_edges_from(graph_edges)
        agent.graph = graph
        attractors = agent.find_attractors()
        self.assertEqual(len(attractors), 2, 'Too many or too few attractors')
        attractors = agent.find_attractors_sources(attractors)
        catchment_area = agent.catchment_area(attractors)
        self.assertEqual(sorted(catchment_area), [11, 14],
                         'Too big or too short catchment area')

        # Two loops attractors
        graph_edges = list()
        for snode, enode in zip(graph_nodes[:11],
                                np.roll(graph_nodes[:11], 1)):
            graph_edges.append((snode, enode))
        for snode, enode in zip(graph_nodes[11:],
                                np.roll(graph_nodes[11:], 1)):
            graph_edges.append((snode, enode))

        graph = nx.DiGraph()
        graph.add_nodes_from(graph_nodes)
        graph.add_edges_from(graph_edges)
        agent.graph = graph
        attractors = agent.find_attractors()
        self.assertEqual(len(attractors), 2, 'Too many or too few attractors')
        attractors = agent.find_attractors_sources(attractors)
        catchment_area = agent.catchment_area(attractors)
        self.assertEqual(sorted(catchment_area), [11, 14],
                         'Too big or too short catchment area')

    def test_neighboring_nodes(self):
        """ Counting neighnoring nodes for 3 situations
        1. Local maxima
        2. Saddle points
        3. Local minima
        """
        agent = naviagent.GraphAgent(self.mydb_filename)
        # Local maxima
        graph_nodes = list(agent.graph.nodes)
        graph_edges = list()
        graph_edges.append((graph_nodes[0],
                            graph_nodes[1]))

        graph = nx.DiGraph()
        graph.add_nodes_from(graph_nodes)
        graph.add_edges_from(graph_edges)
        agent.graph = graph
        neighbors = agent.neighboring_nodes(graph_nodes[0])
        expected_nbh = []
        obtained_nbh = [a for a in neighbors]
        self.assertEqual(sorted(expected_nbh),
                         sorted(obtained_nbh),
                         'Problem neighbors maxima')

        # Saddle points
        graph_edges.append((graph_nodes[1],
                            graph_nodes[2]))

        graph = nx.DiGraph()
        graph.add_nodes_from(graph_nodes)
        graph.add_edges_from(graph_edges)
        agent.graph = graph
        neighbors = agent.neighboring_nodes(graph_nodes[1])
        expected_nbh = [graph_nodes[0]]
        obtained_nbh = [a for a in neighbors]
        self.assertEqual(sorted(expected_nbh),
                         sorted(obtained_nbh),
                         'Problem neighbors saddle')

        # Local maxima points
        graph_edges.append((graph_nodes[3],
                            graph_nodes[2]))

        graph = nx.DiGraph()
        graph.add_nodes_from(graph_nodes)
        graph.add_edges_from(graph_edges)
        agent.graph = graph
        neighbors = agent.neighboring_nodes(graph_nodes[2])
        expected_nbh = [graph_nodes[3], graph_nodes[1]]
        obtained_nbh = [a for a in neighbors]
        self.assertEqual(sorted(expected_nbh),
                         sorted(obtained_nbh),
                         'Problem neighbors minima')


if __name__ == '__main__':
    unittest.main()
